from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

import networkx as nx
import structlog

from pydiverse.pipedag.context import ConfigContext, DAGContext, RunContextServer
from pydiverse.pipedag.errors import DuplicateNameError, FlowError
from pydiverse.pipedag.util.config import PipedagConfig

if TYPE_CHECKING:
    from pydiverse.pipedag.core import Result, Stage, Task
    from pydiverse.pipedag.core.stage import CommitStageTask
    from pydiverse.pipedag.engine import OrchestrationEngine


class Flow:
    def __init__(
        self,
        name: str = "default",
    ):
        self.name = name

        self.logger = structlog.getLogger(module=__name__, cls=self.__class__.__name__)
        self.stages: dict[str, Stage] = {}
        self.tasks: list[Task] = []

        self.graph = nx.DiGraph()
        self.explicit_graph: nx.DiGraph | None = None

    def __enter__(self):
        # Check that flows don't get nested
        try:
            DAGContext.get()
        except LookupError:
            pass
        else:
            raise RuntimeError("DAG Context already exists. Flows can't be nested.")

        # Initialize context (both Flow and Stage use DAGContext to transport
        # information to @materialize annotations within the flow and to
        # support nesting of stages)
        self._ctx = DAGContext(flow=self, stage=None)
        self._ctx.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._ctx.__exit__()
        del self._ctx

        self.explicit_graph = self.build_graph()

    def add_stage(self, stage: Stage):
        if stage.name in self.stages:
            raise DuplicateNameError(f"Stage with name '{stage.name}' already exists.")

        stage.id = len(self.stages)
        self.stages[stage.name] = stage

    def add_task(self, task: Task):
        assert self.stages[task.stage.name] is task.stage

        task.id = len(self.tasks)
        self.tasks.append(task)
        self.graph.add_node(task)

    def add_edge(self, from_: Task, to: Task):
        if from_ not in self.graph:
            raise FlowError(
                f"Can't add edge from {from_} to {to} because `from` is not in the"
                " flow."
            )
        if to not in self.graph:
            raise FlowError(
                f"Can't add edge from {from_} to {to} because `to` is not in the flow."
            )

        self.graph.add_edge(from_, to)

    # noinspection PyPackageRequirements
    def visualize(self):
        # TODO: Also allow visualizing the run result
        #       Successful tasks in green, failed in red, skipped orange
        import pydot

        dot = pydot.Dot()
        subgraphs: dict[Stage, pydot.Subgraph] = {}
        nodes: dict[Task, pydot.Node] = {}

        for stage in self.stages.values():
            s = pydot.Subgraph(
                f"cluster_{stage.name}",
                label=stage.name,
                bgcolor="#00000011",
            )
            subgraphs[stage] = s

            if stage.outer_stage is None:
                dot.add_subgraph(s)
            else:
                subgraphs[stage.outer_stage].add_subgraph(s)

        for task in self.tasks:
            # noinspection PyProtectedMember
            if task._visualize_hidden:
                continue

            node = pydot.Node(
                task.id,
                label=task.name,
                fillcolor="#FFFFFF",
                style="filled",
            )
            nodes[task] = node
            subgraphs[task.stage].add_node(node)

        for nx_edge in self.graph.edges:
            edge = pydot.Edge(nodes[nx_edge[0]], nodes[nx_edge[1]])
            dot.add_edge(edge)

        # Display
        # Either as svg in ipython
        # Or as PDF in default pdf viewer
        try:
            # noinspection PyUnresolvedReferences
            from IPython import get_ipython

            # noinspection PyUnresolvedReferences
            from IPython.display import SVG, display

            ipython = get_ipython()
            if ipython is None or ipython.config.get("IPKernelApp") is None:
                raise RuntimeError(
                    "Either IPython isn't running or SVG aren't supported."
                )

            display(SVG(dot.create_svg()))  # type: ignore
        except (ImportError, RuntimeError):
            import tempfile
            import webbrowser

            f = tempfile.NamedTemporaryFile(suffix=".pdf")
            f.write(dot.create_pdf())  # type: ignore
            webbrowser.open_new("file://" + f.name)

            # Keep alive to prevent immediate deletion of file
            globals()["__pipedag_tmp_file_reference__"] = f

        return dot

    def build_graph(self) -> nx.DiGraph:
        if not nx.is_directed_acyclic_graph(self.graph):
            raise FlowError("Graph is not a DAG")

        explicit_graph = self.graph.copy()
        stages = self.stages.values()

        # Commit Tasks
        commit_tasks: dict[Stage, CommitStageTask] = {}
        for stage in stages:
            commit_tasks[stage] = stage.commit_task
            explicit_graph.add_node(stage.commit_task)

        # Add dependencies
        # Because stages can be nested, we add the commit task dependencies
        # to the leaf nodes of the subgraph
        contained_tasks: defaultdict[Stage, list[Task]] = defaultdict(lambda: [])
        leaf_nodes: defaultdict[Stage, list[Task]] = defaultdict(lambda: [])

        for stage in stages:
            s = stage
            while s is not None:
                contained_tasks[s].extend(stage.tasks)
                s = s.outer_stage

        for stage in stages:
            subgraph = self.graph.subgraph(contained_tasks[stage])  # type: nx.DiGraph
            for task, degree in subgraph.out_degree:  # type: ignore
                if degree != 0:
                    continue
                leaf_nodes[stage].append(task)

        # Add commit task downstream dependencies to leaf nodes
        for stage in stages:
            for leaf in leaf_nodes[stage]:
                commit_task = commit_tasks[stage]
                explicit_graph.add_edge(leaf, commit_task)

        # Add commit task upstream dependencies
        for task in self.tasks:
            for parent, _ in self.graph.in_edges(task):  # type: Task
                if parent.stage == task.stage:
                    continue
                if task.stage.is_inner(parent.stage):
                    continue

                commit_task = commit_tasks[parent.stage]
                explicit_graph.add_edge(commit_task, task)

        # Ensure inner stages get committed before outer stages
        for stage in stages:
            if stage.outer_stage is not None:
                explicit_graph.add_edge(
                    commit_tasks[stage], commit_tasks[stage.outer_stage]
                )

        return explicit_graph

    def run(
        self,
        config_context: ConfigContext = None,
        orchestration_engine: OrchestrationEngine = None,
        fail_fast: bool | None = None,
        ignore_fresh_input: bool = False,
        stages: list[str] = None,
        **kwargs,
    ) -> Result:
        """Execute a flow

        You can provide an engine to execute the flow with using the `engine`
        keyword. If no engine is provided, the engine specified in the config
        file is used.

        :param config_context: A configuration context with information about
            the pipe-DAG instance.
        :param orchestration_engine: The orchestration engine to use.
        :param fail_fast: True means that errors should be raised as exceptions
            out of this function.
        :param ignore_fresh_input: whether cache functions should be disabled that
            check for fresh input into pipe-DAG
        :param kwargs: Other keyword arguments. They get passed on directly to the
            orchestration engine's `.run` method and thus are engine dependant.
        :return:
            Result object that gives information whether run was successful
        """

        # TODO: implement running only a subset of stages (see parameter stages)

        # Get the ConfigContext to use
        if config_context is None:
            try:
                config_context = ConfigContext.get()
            except LookupError:
                config_context = PipedagConfig.default.get()

        with config_context, RunContextServer(self) as run_context:
            # Configure Run Context
            run_context.ignore_fresh_input = ignore_fresh_input

            if orchestration_engine is None:
                orchestration_engine = config_context.create_orchestration_engine()
            result = orchestration_engine.run(flow=self, **kwargs)

        fail_fast = config_context.fail_fast if fail_fast is None else fail_fast
        if not result.successful and fail_fast:
            raise result.exception or Exception("Flow run failed")

        return result



from __future__ import annotations

import functools
import warnings
from typing import TYPE_CHECKING, Callable, Protocol

from pydiverse.pipedag._typing import CallableT

if TYPE_CHECKING:

    class EngineDispatchedFn(Protocol):
        dialect: Callable[[str], Callable[[CallableT], CallableT]]
        original: CallableT


__all__ = [
    "engine_dispatch",
]


def engine_dispatch(fn: CallableT) -> CallableT | EngineDispatchedFn:
    """Dispatch a function based on `self.engine.dialect.name`."""
    fn_variants = {}

    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        self_or_cls = args[0]
        dialect_name = self_or_cls.engine.dialect.name
        dispatched_fn = fn_variants.get(dialect_name, fn)
        return dispatched_fn(*args, **kwargs)

    def dialect_decorator(dialect):
        def decorate(dialect_fn):
            if dialect in fn_variants:
                msg = (
                    f"Already defined a variant of function '{fn.__name__}'"
                    f" for dialect '{dialect}'."
                )
                warnings.warn(msg)
            fn_variants[dialect] = dialect_fn
            return dialect_fn

        return decorate

    wrapped.dialect = dialect_decorator
    wrapped.original = fn
    return wrapped



from __future__ import annotations

import atexit
import os
import threading
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Union

import structlog

from pydiverse.pipedag.context.context import ConfigContext
from pydiverse.pipedag.core.stage import Stage
from pydiverse.pipedag.errors import DisposedError, LockError
from pydiverse.pipedag.util import Disposable, normalize_name, requires

__all__ = [
    "BaseLockManager",
    "LockState",
    "NoLockManager",
    "FileLockManager",
    "ZooKeeperLockManager",
]


class LockState(Enum):
    """Lock State

    Represent the current state of a lock.

    UNLOCKED:
        The lock manager hasn't acquired this lock and a different process
        might be accessing the resource.

    LOCKED:
        The lock manager has acquired the lock, and it hasn't expired.

    UNCERTAIN:
        The lock manager isn't certain about the state of the lock. This
        can, for example, happen if a lock manager requires a connection to
        the internet and this connection is interrupted.
        Any process that depends on the locked resource should assume that
        the resource isn't locked anymore and pause until the lock transitions
        to a different state.

    INVALID:
        The lock has been invalidated. This means that a lock that was LOCKED
        has been unlocked for some unexpected reason. A common transition
        is that a lock transitions from `LOCKED -> UNCERTAIN -> INVALID`.
    """

    UNLOCKED = 0
    LOCKED = 1
    UNCERTAIN = 2
    INVALID = 3


Lockable = Union[Stage, str]  # noqa: UP007
LockStateListener = Callable[[Lockable, LockState, LockState], None]


class BaseLockManager(Disposable, ABC):
    """Lock Manager base class

    A lock manager is responsible for acquiring and releasing locks on
    stage. This is necessary to prevent two flows from accessing the
    same stage at the same time (which would lead to corrupted data).
    """

    def __init__(self):
        self.logger = structlog.get_logger(lock=type(self).__name__)

        self.state_listeners = set()
        self.lock_states = defaultdict(lambda: LockState.UNLOCKED)
        self.__lock_state_lock = threading.Lock()

    @contextmanager
    def __call__(self, lock: Lockable):
        self.acquire(lock)
        try:
            yield
        finally:
            self.release(lock)

    def dispose(self):
        self.release_all()
        super().dispose()

    @abstractmethod
    def acquire(self, lock: Lockable):
        """Acquires a lock to access a given object"""

    @abstractmethod
    def release(self, lock: Lockable):
        """Releases a previously acquired lock"""

    def release_all(self):
        """Releases all aquired locks"""
        locks = list(self.lock_states.items())
        for lock, _state in locks:
            self.release(lock)

    def add_lock_state_listener(self, listener: LockStateListener):
        """Add a function to be called when the state of a lock changes

        The listener will be called with the affected stage, the old lock
        state and the new lock state as arguments.
        """
        if listener is None or not callable(listener):
            raise ValueError("Listener must be callable.")
        self.state_listeners.add(listener)

    def remove_lock_state_listener(self, listener: LockStateListener):
        """Removes a function from the set of listeners"""
        self.state_listeners.remove(listener)

    def set_lock_state(self, lock: Lockable, new_state: LockState):
        """Update the state of a lock

        Function used by lock implementations to update the state of a
        lock. If appropriate, listeners will be informed about this change.
        """
        with self.__lock_state_lock:
            if lock not in self.lock_states:
                self.lock_states[lock] = new_state
                for listener in self.state_listeners:
                    listener(lock, LockState.UNLOCKED, new_state)
            else:
                old_state = self.lock_states[lock]
                self.lock_states[lock] = new_state
                if old_state != new_state:
                    for listener in self.state_listeners:
                        listener(lock, old_state, new_state)

            if new_state == LockState.UNLOCKED:
                del self.lock_states[lock]

    def get_lock_state(self, lock: Lockable) -> LockState:
        """Returns the state of a lock"""
        with self.__lock_state_lock:
            return self.lock_states[lock]


class NoLockManager(BaseLockManager):
    """Non locking lock manager (oxymoron)

    This lock manager doesn't do any locking and only serves as a placeholder
    for an actual lock manager for testing something locally.

    .. WARNING::
        DON'T USE THIS IN A PRODUCTION ENVIRONMENT. A LOCK MANAGER IS
        ESSENTIAL TO PREVENT DATA CORRUPTION.
    """

    def acquire(self, lock: Lockable):
        self.set_lock_state(lock, LockState.LOCKED)

    def release(self, lock: Lockable):
        self.set_lock_state(lock, LockState.UNLOCKED)


try:
    import filelock as fl
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    fl = None


@requires(fl, ImportError("FileLockManager requires 'filelock' to be installed."))
class FileLockManager(BaseLockManager):
    """Lock manager that uses lock files

    For details on how exactly the file locking is implemented, check out the
    `filelock documentation`_.

    .. _`filelock documentation`:
        https://py-filelock.readthedocs.io/en/latest/index.html
    """

    def __init__(self, base_path: str | Path):
        super().__init__()
        self.base_path = Path(base_path).absolute()
        self.locks: dict[Lockable, fl.BaseFileLock] = {}

        os.makedirs(self.base_path, exist_ok=True)

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        instance_id = normalize_name(ConfigContext.get().instance_id)
        base_path = Path(config["base_path"]) / instance_id
        return cls(base_path)

    def acquire(self, lock: Lockable):
        if lock not in self.locks:
            lock_path = self.lock_path(lock)
            self.locks[lock] = fl.FileLock(lock_path)

        f_lock = self.locks[lock]
        if not f_lock.is_locked:
            self.logger.info(f"Locking '{lock}'")
        f_lock.acquire()
        self.set_lock_state(lock, LockState.LOCKED)

    def release(self, lock: Lockable):
        if lock not in self.locks:
            raise LockError(f"No lock '{lock}' found.")

        f_lock = self.locks[lock]
        f_lock.release()

        if not f_lock.is_locked:
            self.logger.info(f"Unlocking '{lock}'")
            os.remove(f_lock.lock_file)
            del self.locks[lock]
            self.set_lock_state(lock, LockState.UNLOCKED)

    def lock_path(self, lock: Lockable) -> Path:
        if isinstance(lock, Stage):
            return self.base_path / (lock.name + ".lock")
        elif isinstance(lock, str):
            return self.base_path / (lock + ".lock")
        else:
            raise NotImplementedError(
                f"Can't lock object of type '{type(lock).__name__}'"
            )


try:
    import kazoo
    from kazoo.client import KazooClient, KazooState
    from kazoo.recipe.lock import Lock as KazooLock
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    kazoo = None


@requires(kazoo, ImportError("ZooKeeperLockManager requires 'kazoo' to be installed."))
class ZooKeeperLockManager(BaseLockManager):
    """Apache ZooKeeper based lock manager

    Uses Apache ZooKeeper to establish `fully distributed locks that are
    globally synchronous` [1]_. The advantage of this approach is that we
    it is highly reliable and that in case our flow crashes, the acquired
    locks automatically get released (the locks are ephemeral).

    .. [1] https://zookeeper.apache.org/doc/r3.1.2/recipes.html#sc_recipes_Locks
    """

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        client = KazooClient(**config)
        instance_id = normalize_name(ConfigContext.get().instance_id)
        base_path = f"/pipedag/locks/{instance_id}/"
        return cls(client, base_path)

    def __init__(self, client: KazooClient, base_path: str):
        super().__init__()

        self.client = client
        self.base_path = base_path
        if not self.client.connected:
            self.client.start()
            atexit.register(self.__atexit)
        self.client.add_listener(self._lock_listener)

        self.locks: dict[Lockable, KazooLock] = {}

    def __atexit(self):
        try:
            self.dispose()
        except DisposedError:
            pass

    def dispose(self):
        self.release_all()
        self.client.stop()
        self.client.close()
        self.lock_states.clear()
        self.locks.clear()
        super().dispose()

    def acquire(self, lock: Lockable):
        zk_lock = self.client.Lock(self.lock_path(lock))
        self.logger.info(f"Locking '{lock}'")
        if not zk_lock.acquire():
            raise LockError(f"Failed to acquire lock '{lock}'")
        self.locks[lock] = zk_lock
        self.set_lock_state(lock, LockState.LOCKED)

    def release(self, lock: Lockable):
        if lock not in self.locks:
            raise LockError(f"No lock '{lock}' found.")

        self.logger.info(f"Unlocking '{lock}'")
        self.locks[lock].release()
        del self.locks[lock]
        self.set_lock_state(lock, LockState.UNLOCKED)

    def lock_path(self, lock: Lockable):
        if isinstance(lock, Stage):
            return self.base_path + lock.name
        elif isinstance(lock, str):
            return self.base_path + lock
        else:
            raise NotImplementedError(
                f"Can't lock object of type '{type(lock).__name__}'"
            )

    def _lock_listener(self, state):
        if state == KazooState.SUSPENDED:
            for lock in self.locks.keys():
                self.set_lock_state(lock, LockState.UNCERTAIN)
        elif state == KazooState.LOST:
            for lock in self.locks.keys():
                self.set_lock_state(lock, LockState.INVALID)
        elif state == KazooState.CONNECTED:
            for lock in self.locks.keys():
                self.set_lock_state(lock, LockState.LOCKED)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from pydiverse.pipedag.util import Disposable

if TYPE_CHECKING:
    from pydiverse.pipedag.core import Flow, Result


class OrchestrationEngine(Disposable, ABC):
    """Flow orchestration engine base class"""

    @abstractmethod
    def run(self, flow: Flow, **kwargs) -> Result:
        """Execute a flow

        :param flow: the pipedag flow to execute
        :param kwargs: Optional keyword arguments. How they get used is
            engine specific.
        :return: A result instance wrapping the flow execution result.
        """



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pydiverse',
 'pydiverse.pipedag',
 'pydiverse.pipedag.backend',
 'pydiverse.pipedag.backend.table',
 'pydiverse.pipedag.backend.table.util',
 'pydiverse.pipedag.context',
 'pydiverse.pipedag.core',
 'pydiverse.pipedag.engine',
 'pydiverse.pipedag.errors',
 'pydiverse.pipedag.materialize',
 'pydiverse.pipedag.materialize.util',
 'pydiverse.pipedag.util']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.39,<2.0.0',
 'attrs>=22.1.0',
 'contextvars>=2.4',
 'msgpack>=1.0.4',
 'networkx>=2.8',
 'packaging>=21.3',
 'pandas>=1.4.3,<2.0.0',
 'pynng>=0.7.1',
 'python-box>=6.1.0',
 'structlog>=22.1.0',
 'tomli>=2.0.1',
 'typing-extensions>=4.1.0,<5']

extras_require = \
{'docs': ['Sphinx>=5.1.1',
          'sphinx-rtd-theme>=1.0.0',
          'sphinxcontrib-apidoc>=0.3.0'],
 'filelock': ['filelock>=3.7.1'],
 'ibm-db2': ['ibm-db>=3.1.4', 'ibm-db-sa>=0.3.8'],
 'mssql': ['pyodbc>=4.0.35', 'pytsql>=1.1.4'],
 'prefect': ['prefect>=1.3,<2'],
 'zookeeper': ['kazoo>=2.8.0']}

setup_kwargs = {
    'name': 'pydiverse-pipedag',
    'version': '0.2.0',
    'description': 'A pipeline orchestration library executing tasks within one python session. It takes care of SQL table (de)materialization, caching and cache invalidation. Blob storage is supported as well for example for storing model files.',
    'long_description': '# pydiverse.pipedag\n\n[![CI](https://github.com/pydiverse/pydiverse.pipedag/actions/workflows/ci.yml/badge.svg)](https://github.com/pydiverse/pydiverse.pipedag/actions/workflows/ci.yml)\n\nA pipeline orchestration library executing tasks within one python session. It takes care of SQL table\n(de)materialization, caching and cache invalidation. Blob storage is supported as well for example\nfor storing model files.\n\nThis is an early stage version 0.x which lacks documentation. Please contact\nhttps://github.com/orgs/pydiverse/teams/code-owners if you like to become an early adopter\nor to contribute early stage usage examples.\n\n## Usage\n\npydiverse.pipedag can either be installed via pypi with `pip install pydiverse-pipedag` or via conda-forge\nwith `conda install pydiverse-pipedag -c conda-forge`.\n\n## Example\n\nA flow can look like this (i.e. put this in a file named `run_pipeline.py`):\n\n```python\nfrom pydiverse.pipedag import materialize, Table, Flow, Stage\nimport sqlalchemy as sa\nimport pandas as pd\n\nfrom pydiverse.pipedag.context import StageLockContext, RunContext\nfrom pydiverse.pipedag.util import setup_structlog\n\n\n@materialize(lazy=True)\ndef lazy_task_1():\n    return sa.select([sa.literal(1).label("x"), sa.literal(2).label("y")])\n\n\n@materialize(lazy=True, input_type=sa.Table)\ndef lazy_task_2(input1: sa.Table, input2: sa.Table):\n    query = sa.select([(input1.c.x * 5).label("x5"), input2.c.a]).select_from(\n        input1.outerjoin(input2, input2.c.x == input1.c.x)\n    )\n    return Table(query, name="task_2_out", primary_key=["a"])\n\n\n@materialize(lazy=True, input_type=sa.Table)\ndef lazy_task_3(input: sa.Table, my_stage: Stage):\n    return sa.text(f"SELECT * FROM {my_stage.transaction_name}.{input.name}")\n\n\n@materialize(lazy=True, input_type=sa.Table)\ndef lazy_task_4(input: sa.Table, prev_stage: Stage):\n    return sa.text(f"SELECT * FROM {prev_stage.name}.{input.name}")\n\n\n@materialize(nout=2, version="1.0.0")\ndef eager_inputs():\n    dfA = pd.DataFrame(\n        {\n            "a": [0, 1, 2, 4],\n            "b": [9, 8, 7, 6],\n        }\n    )\n    dfB = pd.DataFrame(\n        {\n            "a": [2, 1, 0, 1],\n            "x": [1, 1, 2, 2],\n        }\n    )\n    return Table(dfA, "dfA"), Table(dfB, "dfB_%%")\n\n\n@materialize(version="1.0.0", input_type=pd.DataFrame)\ndef eager_task(tbl1: pd.DataFrame, tbl2: pd.DataFrame):\n    return tbl1.merge(tbl2, on="x")\n\n\ndef main():\n    with Flow() as f:\n        with Stage("stage_1"):\n            lazy_1 = lazy_task_1()\n            a, b = eager_inputs()\n\n        with Stage("stage_2") as stage2:\n            lazy_2 = lazy_task_2(lazy_1, b)\n            lazy_3 = lazy_task_3(lazy_2, stage2)\n            eager = eager_task(lazy_1, b)\n\n        with Stage("stage_3"):\n            lazy_4 = lazy_task_4(lazy_2, stage2)\n        _ = lazy_3, lazy_4, eager  # unused terminal output tables\n\n    # Run flow\n    result = f.run()\n    assert result.successful\n\n    # Run in a different way for testing\n    with StageLockContext():\n        result = f.run()\n        assert result.successful\n        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1\n\n\nif __name__ == "__main__":\n    # initialize logging\n    setup_structlog()\n    main()\n```\n\nCreate a file called `pipedag.yaml` in the same directory:\n\n```yaml\nname: pipedag_tests\ntable_store_connections:\n  postgres:\n    # Postgres: this can be used after running `docker-compose up`\n    url: "postgresql://{$POSTGRES_USERNAME}:{$POSTGRES_PASSWORD}@127.0.0.1:6543/{instance_id}"\n\ninstances:\n  __any__:\n    # listen-interface for pipedag context server which synchronizes some task state during DAG execution\n    network_interface: "127.0.0.1"\n    # classes to be materialized to table store even without pipedag Table wrapper (we have loose coupling between\n    # pipedag and pydiverse.transform, so consider adding \'pydiverse.transform.Table\' in your config)\n    auto_table: ["pandas.DataFrame", "sqlalchemy.sql.elements.TextClause", "sqlalchemy.sql.selectable.Selectable"]\n    fail_fast: true\n\n    instance_id: pipedag_default\n    table_store:\n      class: "pydiverse.pipedag.backend.table.SQLTableStore"\n\n      # Postgres: this can be used after running `docker-compose up`\n      table_store_connection: postgres\n      create_database_if_not_exists: True\n\n      # print select statements before being encapsualted in materialize expressions and tables before writing to\n      # database\n      print_materialize: true\n      # print final sql statements\n      print_sql: true\n\n    blob_store:\n      class: "pydiverse.pipedag.backend.blob.FileBlobStore"\n      base_path: "/tmp/pipedag/blobs"\n\n    lock_manager:\n      class: "pydiverse.pipedag.backend.lock.ZooKeeperLockManager"\n      hosts: "localhost:2181"\n\n    orchestration:\n      class: "pydiverse.pipedag.engine.SequentialEngine"\n```\n\nIf you don\'t have a postgres, Microsoft SQL Server, or IBM DB2 database at hand, you can\nstart a postgres database with the following `docker-compose.yaml` file:\n\n```yaml\nversion: "3.9"\nservices:\n  postgres:\n    image: postgres\n    environment:\n      POSTGRES_USER: sa\n      POSTGRES_PASSWORD: Pydiverse23\n      POSTGRES_PORT: 6543\n    ports:\n      - 6543:5432\n  zoo:\n    image: zookeeper\n    environment:\n      ZOO_4LW_COMMANDS_WHITELIST: ruok\n    ports:\n      - 2181:2181\n```\n\nRun `docker-compose up` in the directory of your `docker-compose.yaml` and then execute\nthe flow script as follows with a shell like `bash` and a python environment that\nincludes `pydiverse-pipedag`, `pandas`, and `sqlalchemy`:\n\n```bash\nexport POSTGRES_USERNAME=sa\nexport POSTGRES_PASSWORD=Pydiverse23\npython run_pipeline.py\n```\n\nFinally, you may connect to your localhost postgres database `pipedag_default` and\nlook at tables in schemas `stage_1`..`stage_3`.\n\nIf you don\'t have a SQL UI at hand, you may use `psql` command line tool inside the docker container.\nCheck out the `NAMES` column in `docker ps` output. If the name of your postgres container is\n`example_postgres_1`, then you can look at output tables like this:\n\n```bash\ndocker exec example_postgres_1 psql --username=sa --dbname=pipedag_default -c \'select * from stage_1.dfa;\'\n```\n\nOr more interactively:\n\n```bash\ndocker exec -t -i example_postgres_1 bash\npsql --username=sa --dbname=pipedag_default\n\\dt stage_*.*\nselect * from stage_2.task_2_out;\n```\n\n## Troubleshooting\n\n### Installing mssql odbc driver for linux\n\nInstalling with\ninstructions [here](https://docs.microsoft.com/en-us/sql/connect/odbc/linux-mac/installing-the-microsoft-odbc-driver-for-sql-server?view=sql-server-ver16#suse18)\nworked.\nBut `odbcinst -j` revealed that it installed the configuration in `/etc/unixODBC/*`. But conda installed pyodbc brings\nits own `odbcinst` executable and that shows odbc config files are expected in `/etc/*`. Symlinks were enough to fix the\nproblem. Try `python -c \'import pyodbc;print(pyodbc.drivers())\'` and see whether you get more than an empty list.\nFurthermore, make sure you use 127.0.0.1 instead of localhost. It seems that /etc/hosts is ignored.\n',
    'author': 'QuantCo, Inc.',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
